-- *****************************************************************************
-- *  Copyright (c) 2014 by Michael Fischer (www.emb4fun.de) 
-- *  All rights reserved.
-- *
-- *  Redistribution and use in source and binary forms, with or without 
-- *  modification, are permitted provided that the following conditions 
-- *  are met:
-- *  
-- *  1. Redistributions of source code must retain the above copyright 
-- *     notice, this list of conditions and the following disclaimer.
-- *  2. Redistributions in binary form must reproduce the above copyright
-- *     notice, this list of conditions and the following disclaimer in the 
-- *     documentation and/or other materials provided with the distribution.
-- *  3. Neither the name of the author nor the names of its contributors may 
-- *     be used to endorse or promote products derived from this software 
-- *     without specific prior written permission.
-- *
-- *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
-- *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
-- *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
-- *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
-- *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
-- *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
-- *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
-- *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
-- *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
-- *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
-- *  SUCH DAMAGE.
-- *
-- *****************************************************************************
-- *  History:
-- *
-- *  14.09.2014  mifi  First version
-- *****************************************************************************

-- *****************************************************************************
-- *  DEFINE: Library                                                          *
-- *****************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.jtag_pkg.ALL;


-- *****************************************************************************
-- *  DEFINE: Entity                                                           *
-- *****************************************************************************

entity jtag_tap is
   port ( 
      --  JTAG (IEEE 1149.1) TAP signals
      tck_i      : in  std_logic := '0';  -- Test Clock
      tms_i      : in  std_logic := '0';  -- Test Mode Select
      tdi_i      : in  std_logic := '0';  -- Test Data In
      tdo_o      : out std_logic;         -- Test Data Out
      trst_i     : in  std_logic := '0';  -- Test Reset, optional
      rtck_o     : out std_logic;         -- Return Test Clock, optional
      
      -- For debug purpose only
      state_o    : out std_logic_vector(3 downto 0);
      ir_reg_o   : out std_logic_vector(3 downto 0);
      
      -- Application JTAG signals
      jtck_o     : out std_logic;
      jtdi_o     : out std_logic;
      jtdo1_i    : in  std_logic := '0';
      jtdo2_i    : in  std_logic := '0';
      juser1_o   : out std_logic;
      juser2_o   : out std_logic;
      jcapture_o : out std_logic;
      jshift_o   : out std_logic;
      jupdate_o  : out std_logic;
      jreset_o   : out std_logic
   );
end entity jtag_tap;


-- *****************************************************************************
-- *  DEFINE: Architecture                                                     *
-- *****************************************************************************

architecture syn of jtag_tap is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------
   constant IDCODE    : std_logic_vector(31 downto 0) := x"6D696669";
   
   constant IR_USER1  : std_logic_vector(3 downto 0)  := "1010";
   constant IR_USER2  : std_logic_vector(3 downto 0)  := "1011";
   constant IR_USER3  : std_logic_vector(3 downto 0)  := "1100";  -- Not implemented yet
   constant IR_USER4  : std_logic_vector(3 downto 0)  := "1101";  -- Not implemented yet
   constant IR_IDCODE : std_logic_vector(3 downto 0)  := "1110";
   constant IR_BYPASS : std_logic_vector(3 downto 0)  := "1111";
   
   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   component jtag_tap_fsm is
      port ( 
         --  JTAG (IEEE 1149.1) TAP signals
         tck_i   : in  std_logic := '0';  -- Test Clock
         tms_i   : in  std_logic := '0';  -- Test Mode Select
         trst_i  : in  std_logic := '0';  -- Test Reset, optional
         state_o : out jtag_tap_state_t
      );
   end component jtag_tap_fsm;
   
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------

   --
   --   TEST_LOGIC_RESET, RUN_TEST_IDLE,
   --   SELECT_DR_SCAN, CAPTURE_DR, SHIFT_DR, EXIT1_DR, PAUSE_DR, EXIT2_DR, UPDATE_DR,
   --   SELECT_IR_SCAN, CAPTURE_IR, SHIFT_IR, EXIT1_IR, PAUSE_IR, EXIT2_IR, UPDATE_IR
   --
   signal tap_state       : jtag_tap_state_t := TEST_LOGIC_RESET;

   --
   -- JTAG Debug Port instruction Register
   --   
   signal ir_reg          : std_logic_vector(3 downto 0);
   signal shift_ir_reg    : std_logic_vector(3 downto 0);
   signal debug_ir_reg    : std_logic_vector(3 downto 0);
   
   --
   -- Working signals
   --
   signal data_32bit_reg  : std_logic_vector(31 downto 0);
   signal data_bypass_reg : std_logic := '0';
   signal user            : std_logic := '0'; 
   signal user1           : std_logic := '0';
   signal user2           : std_logic := '0';
   signal ir_tdo          : std_logic := '0';
   signal dr_tdo          : std_logic := '0';
   signal state           : std_logic_vector(3 downto 0) := (others => '0');

begin

   ------------------------------------
   -- TAP Controller
   ------------------------------------
   
   inst_jtag_tap_fsm : jtag_tap_fsm
      port map ( 
         tck_i   => tck_i,
         tms_i   => tms_i,
         trst_i  => trst_i,
         state_o => tap_state
      );

   
   ------------------------------------
   -- TAP Controller - Input 
   ------------------------------------
   
   process (tck_i, trst_i)
   begin
      if (trst_i = '1') then
         ir_reg       <= IR_IDCODE;
         debug_ir_reg <= IR_IDCODE;
         
      elsif rising_edge(tck_i) then
      
         case (tap_state) is
            
            when TEST_LOGIC_RESET =>
               ir_reg       <= IR_IDCODE;
               debug_ir_reg <= IR_IDCODE;

            -- ********************************************

            when CAPTURE_DR =>
               case (ir_reg) is
                  when IR_IDCODE =>
                     data_32bit_reg <= IDCODE;

                  when others =>   
                     -- Do nothing
               end case;
               
            when SHIFT_DR =>
               case (ir_reg) is
                  when IR_IDCODE =>
                     data_32bit_reg <= tdi_i & data_32bit_reg(31 downto 1);
                     
                  when IR_BYPASS =>   
                     data_bypass_reg <= tdi_i;   
                     
                  when others =>
                     -- Do nothing
               end case;
               
            when UPDATE_DR =>   
               case (ir_reg) is
                  
                  when others =>
                     -- Do nothing
               end case;
               
            -- ********************************************
            
            when CAPTURE_IR =>   
               shift_ir_reg <= "0001";

            when SHIFT_IR =>   
               shift_ir_reg <= tdi_i & shift_ir_reg(3 downto 1);
               
            when UPDATE_IR =>
               user1        <= '0';    
               user2        <= '0';    
               debug_ir_reg <= shift_ir_reg;
               
               -- Check for valid instruction
               case (shift_ir_reg) is
                  when IR_USER1 =>
                     user1  <= '1';
                     ir_reg <= shift_ir_reg;
               
                  when IR_USER2 =>
                     user2  <= '1';
                     ir_reg <= shift_ir_reg;
               
                  when IR_IDCODE =>
                     ir_reg <= shift_ir_reg;
                  
                  when IR_BYPASS =>
                     ir_reg <= shift_ir_reg;
                     
                  when others =>
                     -- In case of invalid shift_ir_reg, use bypass register   
                     ir_reg <= IR_BYPASS;
               end case;
               
            -- ********************************************
            
            when others =>
               -- Do nothing here
         end case;
      end if;
   end process;   -- TAP Controller - Input 

   
   ------------------------------------
   -- TAP Controller - Output
   ------------------------------------
   
   process (tck_i, trst_i)
   begin
      if (trst_i = '1') then
         ir_tdo <= '0';
         dr_tdo <= '0';
         
      elsif falling_edge(tck_i) then
      
         case (tap_state) is
            
            when SHIFT_IR =>
               ir_tdo <= shift_ir_reg(0);      
               
            when SHIFT_DR =>   
               case (ir_reg) is
               
                  when IR_IDCODE =>
                     dr_tdo <= data_32bit_reg(0);
                  
                  when IR_BYPASS =>   
                     dr_tdo <= data_bypass_reg; 
               
                  when others =>
                     -- Do nothing here
               end case;
               
            when others =>
               -- Do nothing here
         end case;
      end if;
   end process;   -- TAP Controller - Outputs
   
   
   --
   -- For debug purpose only
   --
   state <= "0000" when (TEST_LOGIC_RESET = tap_state) else
            "0001" when (RUN_TEST_IDLE    = tap_state) else 
            
            "0010" when (SELECT_DR_SCAN   = tap_state) else 
            "0011" when (CAPTURE_DR       = tap_state) else 
            "0100" when (SHIFT_DR         = tap_state) else 
            "0101" when (EXIT1_DR         = tap_state) else 
            "0110" when (PAUSE_DR         = tap_state) else 
            "0111" when (EXIT2_DR         = tap_state) else 
            "1000" when (UPDATE_DR        = tap_state) else 
            
            "1001" when (SELECT_IR_SCAN   = tap_state) else 
            "1010" when (CAPTURE_IR       = tap_state) else 
            "1011" when (SHIFT_IR         = tap_state) else 
            "1100" when (EXIT1_IR         = tap_state) else 
            "1101" when (PAUSE_IR         = tap_state) else 
            "1110" when (EXIT2_IR         = tap_state) else 
            "1111" when (UPDATE_IR        = tap_state) else "0000";


   --------------------------------------------------------
   -- Output
   --------------------------------------------------------
   
   -- JTAG (IEEE 1149.1) signals
   user   <= user1 OR user2;
   tdo_o  <= ir_tdo  when  (SHIFT_IR = tap_state) else
             dr_tdo  when ((SHIFT_DR = tap_state) AND (user  = '0')) else
             jtdo1_i when ((SHIFT_DR = tap_state) AND (user1 = '1')) else
             jtdo2_i when ((SHIFT_DR = tap_state) AND (user2 = '1')) else '0';
     
   rtck_o <= tck_i;

   -- For debug purpose only
   ir_reg_o <= debug_ir_reg;
   state_o  <= state;
   
   -- Application JTAG signals

   jtck_o     <= tck_i;
   jtdi_o     <= tdi_i;
   juser1_o   <= user1;
   juser2_o   <= user2;
   jcapture_o <= '1' when (CAPTURE_DR       = tap_state) else '0';
   jshift_o   <= '1' when (SHIFT_DR         = tap_state) else '0';
   jupdate_o  <= '1' when (UPDATE_DR        = tap_state) else '0';
   jreset_o   <= '1' when (TEST_LOGIC_RESET = tap_state) else '0';
   
end architecture syn;

-- *** EOF ***

