-- ****************************************************************************
-- *  Copyright (c) 2014 by Michael Fischer (www.emb4fun.de)
-- *  All rights reserved.
-- *
-- *  Redistribution and use in source and binary forms, with or without 
-- *  modification, are permitted provided that the following conditions 
-- *  are met:
-- *  
-- *  1. Redistributions of source code must retain the above copyright 
-- *     notice, this list of conditions and the following disclaimer.
-- *  2. Redistributions in binary form must reproduce the above copyright
-- *     notice, this list of conditions and the following disclaimer in the 
-- *     documentation and/or other materials provided with the distribution.
-- *  3. Neither the name of the author nor the names of its contributors may 
-- *     be used to endorse or promote products derived from this software 
-- *     without specific prior written permission.
-- *
-- *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
-- *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
-- *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
-- *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
-- *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
-- *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
-- *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
-- *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
-- *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
-- *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
-- *  SUCH DAMAGE.
-- *
-- ****************************************************************************
-- *  History:
-- *
-- *  14.09.2014  mifi  First version
-- ****************************************************************************

-- ****************************************************************************
-- *  Library                                                                 *
-- ****************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


-- ****************************************************************************
-- *  Entity                                                                  *
-- ****************************************************************************

entity top is
   port ( 
      --
      -- Input clock 
      --
      CLOCK_50      : in  std_logic;
      
      --
      -- JTAG TAP
      --
      nTRST_i       : in  std_logic;
      TDI_i         : in  std_logic; 
      TMS_i         : in  std_logic;
      TCK_i         : in  std_logic;
      RTCK_o        : out std_logic;
      TDO_o         : out std_logic;
      nRESET_i      : in  std_logic;
      
      --
      -- LEDs, green, red and heartbeat
      --
      LED_GREEN     : out std_logic_vector(7 downto 0);
      LED_RED       : out std_logic_vector(8 downto 0);
      LED_HEARTBEAT : out std_logic;
      
      --
      -- Switches and keys
      --
      SW            : in    std_logic_vector(9 downto 0);
      KEY           : in    std_logic_vector(3 downto 0);
      
      --
      -- 7-segment display
      --
      SEG7          : out   std_logic_vector(27 downto 0);
      
      --
      -- Logic analyzer
      --
      LA_TRST_o     : out std_logic;
      LA_TDI_o      : out std_logic;
      LA_TMS_o      : out std_logic;
      LA_TCK_o      : out std_logic;
      LA_TDO_o      : out std_logic;
      LA_STATE_o    : out std_logic_vector(3 downto 0);
      LA_IR_REG_o   : out std_logic_vector(3 downto 0)
   );
end entity top;


-- ****************************************************************************
-- *  Architecture                                                            *
-- ****************************************************************************

architecture syn of top is

   --
   -- Define all components which are included here
   --
   
   component pll_sys
      port ( 
         inclk0   : in  std_logic  := '0';
         c0       : out std_logic ;
         c1       : out std_logic ;
         c2       : out std_logic ;
         locked   : out std_logic 
      );
   end component pll_sys;
   
   
   component heartbeat is
      port (  
         clk         : in  std_logic;
         counter_out : out std_logic
      );
   end component heartbeat;
   

   component cdc_flag is
      port ( 
         clkA_i : in  std_logic := '0';
         flag_i : in  std_logic := '0';
         clkB_i : in  std_logic := '0';
         flag_o : out std_logic
      );
   end component cdc_flag;

   
   component seg7_lut is
      port (
         dig_in  : in  std_logic_vector(3 downto 0);
         seg_out : out std_logic_vector(6 downto 0)
      );
   end component seg7_lut;
      
   
   component jtag_tap is
      port ( 
         --  JTAG (IEEE 1149.1) TAP signals
         tck_i      : in  std_logic := '0';  -- Test Clock
         tms_i      : in  std_logic := '0';  -- Test Mode Select
         tdi_i      : in  std_logic := '0';  -- Test Data In
         tdo_o      : out std_logic;         -- Test Data Out
         trst_i     : in  std_logic := '0';  -- Test Reset, optional
         rtck_o     : out std_logic;         -- Return Test Clock, optional
      
         -- For debug purpose only
         state_o    : out std_logic_vector(3 downto 0);
         ir_reg_o   : out std_logic_vector(3 downto 0);
         
         -- Application JTAG signals
         jtck_o     : out std_logic;
         jtdi_o     : out std_logic;
         jtdo1_i    : in  std_logic := '0';
         jtdo2_i    : in  std_logic := '0';
         juser1_o   : out std_logic;
         juser2_o   : out std_logic;
         jcapture_o : out std_logic;
         jshift_o   : out std_logic;
         jupdate_o  : out std_logic;
         jreset_o   : out std_logic
      );
   end component jtag_tap;
   
   
   --
   -- Define all local signals (like static data) here
   --
   signal clk_10           : std_logic;
   signal clk_sys          : std_logic;
   
   -- SEG7   
   signal seg0             : std_logic_vector(6 downto 0) := (others => '1');
   signal seg1             : std_logic_vector(6 downto 0) := (others => '1');
   signal seg2             : std_logic_vector(6 downto 0) := (others => '1');
   signal seg3             : std_logic_vector(6 downto 0) := (others => '1');
   
   -- System clock domain
   signal seg01_reg        : std_logic_vector(7 downto 0) := (others => '0');
   signal seg2_reg         : std_logic_vector(3 downto 0) := (others => '0');
   signal update1          : std_logic;
   signal update2          : std_logic;
   
   -- JTAG clock domain   
   signal jtck             : std_logic;
   signal jtdi             : std_logic;
   signal jtdo1            : std_logic;
   signal jtdo2            : std_logic;
   signal juser1           : std_logic;
   signal juser2           : std_logic;
   signal jcapture         : std_logic;
   signal jshift           : std_logic;
   signal jupdate          : std_logic;
   signal jreset           : std_logic;
   
   signal juser1_shift_reg : std_logic_vector(7 downto 0) := (others => '0');
   signal juser1_data_reg  : std_logic_vector(7 downto 0) := (others => '0');

   signal juser2_shift_reg : std_logic_vector(3 downto 0) := (others => '0');
   signal juser2_data_reg  : std_logic_vector(3 downto 0) := (others => '0');

   -- Working signals
   signal tdo              : std_logic;
   signal state            : std_logic_vector(3 downto 0) := (others => '0');
   signal ir_reg           : std_logic_vector(3 downto 0) := (others => '0');
   
begin

   inst_pll_sys : pll_sys
      port map ( 
         inclk0 => CLOCK_50,
         c0     => clk_sys,
         c1     => open,
         c2     => clk_10,
         locked => open
      );

                                  
   inst_heartbeat : heartbeat 
      port map ( 
         clk         => clk_10,
         counter_out => LED_HEARTBEAT
      );

      
   inst_seg7_seg3 : seg7_lut
      port map (
         dig_in  => ir_reg,
         seg_out => seg3
      );      


   inst_seg7_seg2 : seg7_lut
      port map (
         dig_in  => seg2_reg,
         seg_out => seg2
      );      


   inst_seg7_seg1 : seg7_lut
      port map (
         dig_in  => seg01_reg(7 downto 4),
         seg_out => seg1
      );      


   inst_seg7_seg0 : seg7_lut
      port map (
         dig_in  => seg01_reg(3 downto 0),
         seg_out => seg0
      );      

      
   inst_jtag_tap : jtag_tap
      port map ( 
         --  JTAG (IEEE 1149.1) TAP signals
         tck_i      => TCK_i,
         tms_i      => TMS_i,
         tdi_i      => TDI_i,
         tdo_o      => tdo,
         trst_i     => NOT nTRST_i,
         rtck_o     => RTCK_o,

         -- For debug purpose only
         state_o    => state,
         ir_reg_o   => ir_reg,
         
         -- Application JTAG signals
         jtck_o     => jtck,
         jtdi_o     => jtdi,
         jtdo1_i    => jtdo1,
         jtdo2_i    => jtdo2,
         juser1_o   => juser1,
         juser2_o   => juser2,
         jcapture_o => jcapture,
         jshift_o   => jshift,
         jupdate_o  => jupdate,
         jreset_o   => jreset
      );
      
      
   --
   -- Handle USER register, input.
   -- !!! JTAG clock domain !!!
   --
   process (jtck, jreset)
   begin
      if (jreset = '1') then
         juser1_shift_reg <= (others => '0');
         juser1_data_reg  <= (others => '0');

         juser2_shift_reg <= (others => '0');
         juser2_data_reg  <= (others => '0');
         
      elsif rising_edge(jtck) then
         
         -- Handle USER1 register
         if (juser1 = '1') then   
            if (jcapture = '1') then
               juser1_shift_reg <= SW(7 downto 0);
            end if;
         
            if (jshift = '1') then
               juser1_shift_reg <= jtdi & juser1_shift_reg(7 downto 1);
            end if;
         
            if (jupdate = '1') then
               juser1_data_reg <= juser1_shift_reg;
            end if;
         end if;            

         -- Handle USER2 register
         if (juser2 = '1') then   
            if (jcapture = '1') then
               juser2_shift_reg <= NOT KEY;
            end if;
         
            if (jshift = '1') then
               juser2_shift_reg <= jtdi & juser2_shift_reg(3 downto 1);
            end if;
         
            if (jupdate = '1') then
               juser2_data_reg <= juser2_shift_reg;
            end if;
         end if;            
      
      end if;
   end process;   


   --
   -- Handle USER register, output.
   -- !!! JTAG clock domain !!!
   --
   process (jtck)
   begin
      if falling_edge(jtck) then
         
         -- Handle USER1 register
         if (juser1 = '1') then   
            if (jshift = '1') then
               jtdo1 <= juser1_shift_reg(0);
            end if;
         end if;            
      
         -- Handle USER2 register
         if (juser2 = '1') then   
            if (jshift = '1') then
               jtdo2 <= juser2_shift_reg(0);
            end if;
         end if;            
      
      end if;
   end process; 
   
   
   --
   -- Clock domain crossing between JTAG and system
   --  
   inst_cdc_flag1 : cdc_flag
      port map ( 
         clkA_i => jtck,
         flag_i => (juser1 AND jupdate),
         clkB_i => clk_sys,
         flag_o => update1
      );
      
   inst_cdc_flag2 : cdc_flag
      port map ( 
         clkA_i => jtck,
         flag_i => (juser2 AND jupdate),
         clkB_i => clk_sys,
         flag_o => update2
      );
   
   process (clk_sys)
   begin
      if rising_edge(clk_sys) then
      
         if (update1 = '1') then
            seg01_reg <= juser1_data_reg;
         end if;
         
         if (update2 = '1') then
            seg2_reg <= juser2_data_reg;
         end if;
         
      end if;
   end process; 


   --------------------------------------------------------
   -- Output
   --------------------------------------------------------

   SEG7 <= seg3 & seg2 & seg1 & seg0;

         
   LED_GREEN(0) <= nTRST_i;
   LED_GREEN(1) <= TDI_i;
   LED_GREEN(2) <= TMS_i;
   LED_GREEN(3) <= TCK_i;
   LED_GREEN(4) <= nRESET_i;
   LED_GREEN(5) <= '0';
   LED_GREEN(6) <= '0';
   LED_GREEN(7) <= '0';
   

   LED_RED(8 downto 0) <= (others => '0');
   
   TDO_o <= tdo;

   LA_TRST_o   <= NOT nTRST_i; 
   LA_TDI_o    <= TDI_i;
   LA_TMS_o    <= TMS_i;
   LA_TCK_o    <= TCK_i;
   LA_TDO_o    <= tdo; 
   LA_STATE_o  <= state;  
   LA_IR_REG_o <= ir_reg; 
         
end architecture syn;

-- *** EOF ***

